bl_info = {
    "name": "ARchive_neXt Model Importer",
    "author": "AlexNG",
    "version": (1, 0, 0),
    "blender": (2, 83, 0),
    "location": "File > Import > ARchive_neXt Model",
    "description": "Import Assassin's Creed models, io: mesh",
    "warning": "",
    "wiki_url": "",
    "tracker_url": "",
    "category": "Import-Export"
}

import os
import bpy
import collections
import hashlib
import traceback

from bpy_extras.io_utils import ImportHelper
from bpy_extras.image_utils import load_image

from bpy.props import StringProperty, BoolProperty, EnumProperty, CollectionProperty
from bpy.types import Operator

from struct import unpack, calcsize, iter_unpack
from mathutils import Vector, Color, Euler
from itertools import repeat


def robust_basename(fn):
    fn = fn or ''
    fn = fn.rpartition('/')[2]
    fn = fn.rpartition('\\')[2]
    return fn

def file_as_bytes(filename):
    with open(filename, 'rb') as file:
        return file.read()

def sha1hash(data):
    m = hashlib.sha1()
    m.update(data)
    return m.hexdigest()

def find_in_dir(dirname, filename):
    path = os.path.join(dirname, filename)
    
    if os.path.isfile(path):
        return path
    
    # Try brute force case-insensitive search
    if not os.path.isfile(path):
        for item in os.listdir(dirname):
            if item.lower() == filename.lower():
                return os.path.join(dirname, item)
        
    return None


MeshData = collections.namedtuple('MeshData', [
    'verts', 'uvs', 'faces',
    'material', 'diffuse', 'normal', 'specular',
])

class ArxModel:
    def __init__(self, filename):
        self.filename = filename
    
    def read_bytes(self, size):
        buf = self.input_file.read(size)
        if len(buf) != size:
            self.abort('fatal: end of file')
        return buf
    
    def read(self, spec):
        data = unpack(spec, self.read_bytes(calcsize(spec)))
        if len(data) == 1:
            return data[0]
        else:
            return data
    
    def read_array(self, spec, count):
        size = calcsize(spec)
        buf = self.read_bytes(size * count)
        return list(iter_unpack(spec, buf))
    
    def read_string(self):
        length = self.read('I')
        return self.read_bytes(length).decode("utf-8")
    
    def read_name(self):
        name = self.read_string()
        return None if name in ('NULL', 'NONE') else name
    
    def read_mesh(self, size):
        verts = self.read_array('fff', size[0])
        uvs = self.read_array('ff', size[0])
        
        if self.version in (b'V4', b'V5'):
            ftype = 'HHH'
        else:
            ftype = 'III'

        faces = self.read_array(ftype, size[1])

        material = self.read_name()

        if material is not None:
            diffuse = self.read_name()
            normal = self.read_name()
            specular = self.read_name()
        else:
            diffuse = normal = specular = None
        
        return MeshData(verts, uvs, faces, material, diffuse, normal, specular)
    
    def read_model(self):
        with open(self.filename, 'rb') as self.input_file:
            magic = self.read_bytes(4)
            if magic != b'ARX_':
                raise Exception("Not an ARX file")
        
            self.version = self.read_bytes(2)
            if self.version not in (b'V2', b'V3', b'V4', b'V5'):
                raise Exception("Unknown file version")
        
            if self.version != b'V2':
                self.model_scale = self.read('f')
            else:
                self.model_scale = 1
        
            mesh_sizes = self.read_array('II', self.read('I'))
            
            self.meshes = [ self.read_mesh(sz) for sz in mesh_sizes ]

        return self


class MaterialBuilder:
    def __init__(self, operator, context, dirname):
        self.operator = operator
        self.context = context
        self.dirname = dirname
        self.cache = {}
        self.img_cache = {}
        self.known_groups = {}
        
        for mat in bpy.data.materials:
            if 'arx_name' in mat:
                self.cache[mat['arx_name']] = mat
                
        for img in bpy.data.images:
            if 'arx_hash' in img:
                self.img_cache[img['arx_hash']] = img

        for group in bpy.data.node_groups:
            if 'arx_special' in group:
                self.known_groups[group['arx_special']] = group

    def report(self, sev, message):
        self.operator.report(sev, message)
        print(message)

    def get_group(self, group_name):
        builder = getattr(self, 'build_group_'+group_name)

        group = self.known_groups.get(group_name, None)
        if not group:
            group = self.known_groups[group_name] = builder()
            group['arx_special'] = group_name

        return group

    def build_group_AC_Normals(self):
        # Generated by Nodes To Python
        node_tree = bpy.data.node_groups.new('AC-Normals', 'ShaderNodeTree')
        socket = node_tree.inputs.new('NodeSocketColor','Image')
        socket.default_value = (0.5, 0.5, 1, 1)
        socket = node_tree.inputs.new('NodeSocketFloatUnsigned','Alpha')
        socket.default_value = 1.0
        socket.min_value = 0.0
        socket.max_value = 1.0
        socket = node_tree.outputs.new('NodeSocketColor','Image')

        nodes = {}
        nodes['Group Input'] = node = node_tree.nodes.new('NodeGroupInput')
        node.name = 'Group Input'
        node.location = (-1109, -21)
        nodes['Separate RGB'] = node = node_tree.nodes.new('ShaderNodeSeparateRGB')
        node.name = 'Separate RGB'
        node.location = (-878, 136)
        nodes['Math.002'] = node = node_tree.nodes.new('ShaderNodeMath')
        node.name = 'Math.002'
        node.location = (-661, 24)
        node.operation = 'MULTIPLY'
        nodes['Group Output'] = node = node_tree.nodes.new('NodeGroupOutput')
        node.name = 'Group Output'
        node.location = (946, 36)
        nodes['Combine XYZ.001'] = node = node_tree.nodes.new('ShaderNodeCombineXYZ')
        node.name = 'Combine XYZ.001'
        node.location = (558, 48)
        nodes['Mapping.001'] = node = node_tree.nodes.new('ShaderNodeMapping')
        node.name = 'Mapping.001'
        node.location = (753, 172)
        node.vector_type = 'POINT'
        node.inputs['Location'].default_value = Vector((0.0, 0.0, 0.5))
        node.inputs['Rotation'].default_value = Euler((0.0, 0.0, 0.0), 'XYZ')
        node.inputs['Scale'].default_value = Vector((1.0, 1.0, 0.5))
        nodes['Math'] = node = node_tree.nodes.new('ShaderNodeMath')
        node.name = 'Math'
        node.location = (360, 183)
        node.operation = 'SQRT'
        nodes['Combine XYZ'] = node = node_tree.nodes.new('ShaderNodeCombineXYZ')
        node.name = 'Combine XYZ'
        node.location = (-424, 237)
        node.inputs['Z'].default_value = 0.0
        nodes['Math.001'] = node = node_tree.nodes.new('ShaderNodeMath')
        node.name = 'Math.001'
        node.location = (163, 234)
        node.operation = 'SUBTRACT'
        node.inputs[0].default_value = 1.0
        nodes['Mapping'] = node = node_tree.nodes.new('ShaderNodeMapping')
        node.name = 'Mapping'
        node.location = (-229, 418)
        node.vector_type = 'POINT'
        node.inputs['Location'].default_value = Vector((-1.0, -1.0, 0.0))
        node.inputs['Rotation'].default_value = Euler((0.0, 0.0, 0.0), 'XYZ')
        node.inputs['Scale'].default_value = Vector((2.0, 2.0, 0.0))
        nodes['Vector Math'] = node = node_tree.nodes.new('ShaderNodeVectorMath')
        node.name = 'Vector Math'
        node.location = (-32, 222)
        node.operation = 'DOT_PRODUCT'

        node_tree.links.new(nodes['Group Input'].outputs['Image'],nodes['Separate RGB'].inputs['Image'])
        node_tree.links.new(nodes['Math.001'].outputs['Value'],nodes['Math'].inputs[0])
        node_tree.links.new(nodes['Combine XYZ'].outputs['Vector'],nodes['Mapping'].inputs['Vector'])
        node_tree.links.new(nodes['Separate RGB'].outputs['R'],nodes['Math.002'].inputs[0])
        node_tree.links.new(nodes['Group Input'].outputs['Alpha'],nodes['Math.002'].inputs[1])
        node_tree.links.new(nodes['Vector Math'].outputs['Value'],nodes['Math.001'].inputs[1])
        node_tree.links.new(nodes['Math.002'].outputs['Value'],nodes['Combine XYZ'].inputs['X'])
        node_tree.links.new(nodes['Separate RGB'].outputs['G'],nodes['Combine XYZ'].inputs['Y'])
        node_tree.links.new(nodes['Mapping'].outputs['Vector'],nodes['Vector Math'].inputs[0])
        node_tree.links.new(nodes['Math.002'].outputs['Value'],nodes['Combine XYZ.001'].inputs['X'])
        node_tree.links.new(nodes['Separate RGB'].outputs['G'],nodes['Combine XYZ.001'].inputs['Y'])
        node_tree.links.new(nodes['Math'].outputs['Value'],nodes['Combine XYZ.001'].inputs['Z'])
        node_tree.links.new(nodes['Combine XYZ.001'].outputs['Vector'],nodes['Mapping.001'].inputs['Vector'])
        node_tree.links.new(nodes['Mapping.001'].outputs['Vector'],nodes['Group Output'].inputs['Image'])
        node_tree.links.new(nodes['Mapping'].outputs['Vector'],nodes['Vector Math'].inputs[1])
        return node_tree

    def build_group_AC_Specular(self):
        # Generated by Nodes To Python
        node_tree = bpy.data.node_groups.new('AC-Specular', 'ShaderNodeTree')
        socket = node_tree.inputs.new('NodeSocketColor','Color')
        socket.default_value = (0.1, 0.1, 0.1, 1.0)
        socket = node_tree.outputs.new('NodeSocketFloat','Metallic')
        socket = node_tree.outputs.new('NodeSocketFloat','Specular')
        socket = node_tree.outputs.new('NodeSocketFloat','Roughness')

        nodes = {}
        nodes['Math'] = node = node_tree.nodes.new('ShaderNodeMath')
        node.name = 'Math'
        node.location = (252, 573)
        node.operation = 'MULTIPLY'
        nodes['Reroute'] = node = node_tree.nodes.new('NodeReroute')
        node.name = 'Reroute'
        node.location = (-74, 134)
        nodes['Reroute.001'] = node = node_tree.nodes.new('NodeReroute')
        node.name = 'Reroute.001'
        node.location = (-95, 141)
        nodes['Group Input'] = node = node_tree.nodes.new('NodeGroupInput')
        node.name = 'Group Input'
        node.location = (-508, 39)
        nodes['Map Range'] = node = node_tree.nodes.new('ShaderNodeMapRange')
        node.name = 'Map Range'
        node.location = (-2, 715)
        node.width = 185
        node.label = 'Saturation Metal'
        node.clamp = True
        node.interpolation_type = 'LINEAR'
        node.inputs['From Min'].default_value = 0.04
        node.inputs['From Max'].default_value = 0.5
        node.inputs['To Min'].default_value = 0.0
        node.inputs['To Max'].default_value = 1.0
        node.inputs['Steps'].default_value = 4.0
        nodes['Map Range.002'] = node = node_tree.nodes.new('ShaderNodeMapRange')
        node.name = 'Map Range.002'
        node.location = (-2, 470)
        node.width = 185
        node.label = 'Saturation Weight'
        node.clamp = True
        node.interpolation_type = 'SMOOTHSTEP'
        node.inputs['From Min'].default_value = 0.04
        node.inputs['From Max'].default_value = 0.10
        node.inputs['To Min'].default_value = 0.0
        node.inputs['To Max'].default_value = 1.0
        node.inputs['Steps'].default_value = 4.0
        nodes['Map Range.003'] = node = node_tree.nodes.new('ShaderNodeMapRange')
        node.name = 'Map Range.003'
        node.location = (-2, 249)
        node.width = 185
        node.label = 'Value Metal'
        node.clamp = True
        node.interpolation_type = 'LINEAR'
        node.inputs['From Min'].default_value = 0.1
        node.inputs['From Max'].default_value = 0.4
        node.inputs['To Min'].default_value = 0.0
        node.inputs['To Max'].default_value = 1.0
        node.inputs['Steps'].default_value = 4.0
        nodes['Group Output'] = node = node_tree.nodes.new('NodeGroupOutput')
        node.name = 'Group Output'
        node.location = (948, 32)
        nodes['Separate HSV'] = node = node_tree.nodes.new('ShaderNodeSeparateHSV')
        node.name = 'Separate HSV'
        node.location = (-319, 109)
        nodes['Math.001'] = node = node_tree.nodes.new('ShaderNodeMath')
        node.name = 'Math.001'
        node.location = (470, 347)
        node.operation = 'ADD'
        node.use_clamp = True
        nodes['RGB Curves'] = node = node_tree.nodes.new('ShaderNodeRGBCurve')
        node.name = 'RGB Curves'
        node.location = (-55, -14)
        node.width = 295
        points = node.mapping.curves[0].points
        points[0].handle_type = 'VECTOR'
        points[1].location = (0.075, 0.31)
        points.new(0.28, 0.56)
        points.new(1.0, 0.78)
        points = node.mapping.curves[1].points
        points[0].location = (0.0, 0.65)
        points[0].handle_type = 'VECTOR'
        points[1].location = (0.037, 0.45)
        points.new(0.124, 0.25)
        points.new(0.35, 0.15)
        points[3].handle_type = 'VECTOR'
        points.new(1.0, 0.075)
        points[4].handle_type = 'VECTOR'
        points = node.mapping.curves[2].points
        points[0].location = (0.0, 0.5)
        points[0].handle_type = 'VECTOR'
        points[1].location = (0.25, 0.3)
        points.new(1.0, 0.05)
        points[2].handle_type = 'VECTOR'
        node.mapping.update()
        nodes['Separate RGB'] = node = node_tree.nodes.new('ShaderNodeSeparateRGB')
        node.name = 'Separate RGB'
        node.location = (308, 25)
        nodes['Mix'] = node = node_tree.nodes.new('ShaderNodeMixRGB')
        node.name = 'Mix'
        node.location = (743, -43)
        node.blend_type = 'MIX'

        node_tree.links.new(nodes['Group Input'].outputs['Color'],nodes['Separate HSV'].inputs['Color'])
        node_tree.links.new(nodes['Reroute.001'].outputs['Output'],nodes['Map Range'].inputs['Value'])
        node_tree.links.new(nodes['Separate HSV'].outputs['V'],nodes['RGB Curves'].inputs['Color'])
        node_tree.links.new(nodes['RGB Curves'].outputs['Color'],nodes['Separate RGB'].inputs['Image'])
        node_tree.links.new(nodes['Map Range'].outputs['Result'],nodes['Math'].inputs[0])
        node_tree.links.new(nodes['Reroute'].outputs['Output'],nodes['Map Range.002'].inputs['Value'])
        node_tree.links.new(nodes['Map Range.002'].outputs['Result'],nodes['Math'].inputs[1])
        node_tree.links.new(nodes['Math'].outputs['Value'],nodes['Math.001'].inputs[0])
        node_tree.links.new(nodes['Reroute'].outputs['Output'],nodes['Map Range.003'].inputs['Value'])
        node_tree.links.new(nodes['Map Range.003'].outputs['Result'],nodes['Math.001'].inputs[1])
        node_tree.links.new(nodes['Separate HSV'].outputs['V'],nodes['Reroute'].inputs['Input'])
        node_tree.links.new(nodes['Separate HSV'].outputs['S'],nodes['Reroute.001'].inputs['Input'])
        node_tree.links.new(nodes['Separate RGB'].outputs['G'],nodes['Mix'].inputs['Color1'])
        node_tree.links.new(nodes['Separate RGB'].outputs['B'],nodes['Mix'].inputs['Color2'])
        node_tree.links.new(nodes['Math.001'].outputs['Value'],nodes['Mix'].inputs['Fac'])
        node_tree.links.new(nodes['Mix'].outputs['Color'],nodes['Group Output'].inputs['Roughness'])
        node_tree.links.new(nodes['Separate RGB'].outputs['R'],nodes['Group Output'].inputs['Specular'])
        node_tree.links.new(nodes['Math.001'].outputs['Value'],nodes['Group Output'].inputs['Metallic'])
        return node_tree

    def build_texture(self, tex_name):
        if not tex_name:
            return None

        # Try same directory as arx file
        name = None

        if os.path.isfile(tex_name):
            name = tex_name

        if not name:
            name = find_in_dir(self.dirname, robust_basename(tex_name))

        # Try all subdirectories relative to arx file
        if not name:
            for item in os.listdir(self.dirname):
                subdir = os.path.join(self.dirname, item)
                if os.path.isdir(subdir):
                    name = find_in_dir(subdir, robust_basename(tex_name))
                    if name:
                        break
            
        if not name:
            self.report({'WARNING'}, f'Could not find texture: {tex_name}')
            return None
        
        key = sha1hash(file_as_bytes(name))
        
        try:
            return self.img_cache[key]
        except KeyError:
            pass

        img = load_image(name)

        img['arx_hash'] = key
        self.img_cache[key] = img
        return img

    def build_material_tree(self, mat, info):
        node_tree = mat.node_tree

        # Clear the tree
        for n in list(node_tree.nodes):
            node_tree.nodes.remove(n)

        nodes = {}

        nodes['Material Output'] = node = node_tree.nodes.new('ShaderNodeOutputMaterial')
        node.name = 'Material Output'
        node.location = (300, 300)

        nodes['Principled BSDF'] = node = node_tree.nodes.new('ShaderNodeBsdfPrincipled')
        node.name = 'Principled BSDF'
        node.location = (10, 300)
        node.distribution = 'GGX'
        node.subsurface_method = 'BURLEY'

        node_tree.links.new(nodes['Principled BSDF'].outputs['BSDF'],nodes['Material Output'].inputs['Surface'])

        nodes['UV Map'] = node = node_tree.nodes.new('ShaderNodeUVMap')
        node.name = 'UV Map'
        node.location = (-1130, 98)
        node.from_instancer = False
        node.uv_map = 'uv'

        # If no textures present, the material is obviously incomplete.
        # In this case, generate all nodes and let the user sort it out.
        force = not (info.diffuse or info.specular or info.normal)
        
        if force:
            self.report({'INFO'}, f'Material {info.material} has no textures.')

        if info.diffuse or force:
            nodes['diffuse'] = node = node_tree.nodes.new('ShaderNodeTexImage')
            node.name = 'diffuse'
            node.label = robust_basename(info.diffuse)
            node.location = (-900, 300)
            node.width = 350
            node.image = image = self.build_texture(info.diffuse)

            if image and image.channels == 4 and image.depth not in (8, 24, 96):
                mat.shadow_method = 'CLIP'
                mat.blend_method = 'CLIP'

            node_tree.links.new(nodes['UV Map'].outputs['UV'],nodes['diffuse'].inputs['Vector'])
            node_tree.links.new(nodes['diffuse'].outputs['Color'],nodes['Principled BSDF'].inputs['Base Color'])
            node_tree.links.new(nodes['diffuse'].outputs['Alpha'],nodes['Principled BSDF'].inputs['Alpha'])

        if info.specular or force:
            nodes['specular'] = node = node_tree.nodes.new('ShaderNodeTexImage')
            node.name = 'specular'
            node.location = (-900, 40)
            node.width = 350
            node.label = robust_basename(info.specular)
            node.image = self.build_texture(info.specular)

            nodes['AC_Specular'] = node = node_tree.nodes.new('ShaderNodeGroup')
            node.name = 'AC_Specular'
            node.location = (-478, 34)
            node.width = 200
            node.node_tree = self.get_group('AC_Specular')

            node_tree.links.new(nodes['UV Map'].outputs['UV'],nodes['specular'].inputs['Vector'])
            node_tree.links.new(nodes['specular'].outputs['Color'],nodes['AC_Specular'].inputs['Color'])
            node_tree.links.new(nodes['AC_Specular'].outputs['Metallic'],nodes['Principled BSDF'].inputs['Metallic'])
            node_tree.links.new(nodes['AC_Specular'].outputs['Specular'],nodes['Principled BSDF'].inputs['Specular'])
            node_tree.links.new(nodes['AC_Specular'].outputs['Roughness'],nodes['Principled BSDF'].inputs['Roughness'])

        if info.normal or force:
            nodes['normal'] = node = node_tree.nodes.new('ShaderNodeTexImage')
            node.name = 'normal'
            node.location = (-900, -230)
            node.width = 350
            node.label = robust_basename(info.normal)
            node.image = img = self.build_texture(info.normal)
            
            if img:
                img.colorspace_settings.name = 'Non-Color'
                img.alpha_mode = 'CHANNEL_PACKED'

            nodes['AC_Normals'] = node = node_tree.nodes.new('ShaderNodeGroup')
            node.name = 'AC_Normals'
            node.location = (-480, -230)
            node.width = 200
            node.node_tree = self.get_group('AC_Normals')

            nodes['Normal Map'] = node = node_tree.nodes.new('ShaderNodeNormalMap')
            node.name = 'Normal Map'
            node.location = (-220, -230)
            node.space = 'TANGENT'
            node.uv_map = 'uv'

            node_tree.links.new(nodes['UV Map'].outputs['UV'],nodes['normal'].inputs['Vector'])
            node_tree.links.new(nodes['normal'].outputs['Color'],nodes['AC_Normals'].inputs['Image'])
            node_tree.links.new(nodes['normal'].outputs['Alpha'],nodes['AC_Normals'].inputs['Alpha'])
            node_tree.links.new(nodes['AC_Normals'].outputs['Image'],nodes['Normal Map'].inputs['Color'])
            node_tree.links.new(nodes['Normal Map'].outputs['Normal'],nodes['Principled BSDF'].inputs['Normal'])
        
    def build_material(self, info):
        if not info.material:
            return None

        try:
            return self.cache[info.material]
        except KeyError:
            pass

        mat = bpy.data.materials.new(name=info.material)
        mat.use_nodes = True
        
        self.build_material_tree(mat, info)

        mat['arx_name'] = info.material
        self.cache[info.material] = mat
        return mat


class ModelBuilder:
    def __init__(self, operator, context, dirname, scale):
        self.context = context
        self.mat_builder = MaterialBuilder(operator, context, dirname)
        self.scale = scale
    
    def build_mesh(self, model):
        verts = []
        uvs = []
        tris = []

        mat_map = []
        scale = self.scale
        
        for i, meshinfo in enumerate(model.meshes):
            start_vert = len(verts) - 1

            verts += [
                (x / scale, y / scale, z / scale)
                for x, y, z in meshinfo.verts
            ]

            uvs += [
                (u, 1-v)
                for u,v in meshinfo.uvs
            ]

            tris += [
                (start_vert + a, start_vert + b, start_vert + c)
                for a, b, c in meshinfo.faces
            ]

            mat_map += repeat(i, len(meshinfo.faces))

        assert len(uvs) == len(verts)

        # Create the mesh
        mesh = bpy.data.meshes.new(name=robust_basename(model.filename))

        mesh.from_pydata(verts, [], tris)
        
        # Assign shading
        for f in mesh.polygons:
            f.use_smooth = True
        
        # Assign UV
        loops = mesh.loops

        uv_data = mesh.uv_layers.new(name='uv').data
        for i in range(len(uv_data)):
            uv_data[i].uv = uvs[loops[i].vertex_index]

        # Assign materials
        mesh.materials.clear()

        for meshinfo in model.meshes:
            mat = self.mat_builder.build_material(meshinfo)
            mesh.materials.append(mat)

        mesh.polygons.foreach_set('material_index', mat_map)
        return mesh

    def build_model(self, model):
        mesh = self.build_mesh(model)
        obj = bpy.data.objects.new(name=mesh.name, object_data=mesh)
        obj.scale = (model.model_scale,)*3
        self.context.collection.objects.link(obj)
        return obj


class ImportArchiveNext(Operator, ImportHelper):
    """Import an Assassin's Creed model exported via the ARchive_neXt program."""
    bl_idname = "import_mesh.archive_next"
    bl_label = "Import ARchive_neXt"
    bl_options = {'UNDO'}

    filename_ext = ".arx"
    filter_glob: StringProperty(default="*.arx", options={'HIDDEN'})
    directory: StringProperty(options={'HIDDEN'})
    files: CollectionProperty(name="File Path", type=bpy.types.OperatorFileListElement)

    scale_unit: EnumProperty(
        name="Unit Scale",
        description="Apply a scale factor to coordinates read from the file",
        items=[
            ("S1", "No Scaling", "Import coordinates as is"),
            ("S31460", "1/31460 Scale (AC1, AC2 Characters)", "Used for older characters, some objects"),
            ("S100000000", "1/100000000 Scale (Objects, AC2B Characters)", "Used for objects, Brotherhood characters"),
        ],
        default="S100000000",
    )
        
    def draw(self, context):
        layout = self.layout
        layout.operator('file.select_all')
        layout.label(text="Scale Meshes:")
        layout.prop(self, 'scale_unit', text='')
        layout.label(text="Values based on 2BU character height", icon='INFO')

    def execute(self, context):
        fnames = [f.name for f in self.files]
        models = []
        
        for filename in fnames:
            try:
                model = ArxModel(os.path.join(self.directory, filename))
                models.append(model.read_model())
            except Exception as e:
                self.report({'ERROR'}, f'Could not read file {filename}: {e}')
                traceback.print_exc()
                return {'CANCELLED'}
        
        scale = float(self.scale_unit[1:])
        builder = ModelBuilder(self, context, self.directory, scale)
        
        for model in models:
            try:
                builder.build_model(model)
            except Exception as e:
                self.report({'ERROR'}, f'Could not import {model.filename}: {e}')
                traceback.print_exc()
                return {'FINISHED'}

        return {'FINISHED'}


def menu_func_import(self, context):
    self.layout.operator(ImportArchiveNext.bl_idname, text="ARchive_neXt Model (.arx)")


def register():
    bpy.utils.register_class(ImportArchiveNext)
    bpy.types.TOPBAR_MT_file_import.append(menu_func_import)


def unregister():
    bpy.utils.unregister_class(ImportArchiveNext)
    bpy.types.TOPBAR_MT_file_import.remove(menu_func_import)


if __name__ == "__main__":
    register()
